@@ -440,6 +440,24 @@ class SessionTests(utils.TestCase):
         self.assertThat(self.requests_mock.request_history,
                         matchers.HasLength(retries + 1))
 
+    def test_connect_retries_interval_limit(self):
+        self.stub_url('GET', exc=requests.exceptions.Timeout())
+
+        session = client_session.Session()
+        retries = 20
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.ConnectTimeout,
+                              session.get,
+                              self.TEST_URL, connect_retries=retries)
+
+            self.assertEqual(retries, m.call_count)
+            # The interval maxes out at 60
+            m.assert_called_with(60.0)
+
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
     def test_http_503_retries(self):
         self.stub_url('GET', status_code=503)
 
@@ -496,6 +514,26 @@ class SessionTests(utils.TestCase):
         self.assertThat(self.requests_mock.request_history,
                         matchers.HasLength(1))
 
+    def test_http_status_retries_inverval_limit(self):
+        self.stub_url('GET', status_code=409)
+
+        session = client_session.Session()
+        retries = 20
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.Conflict,
+                              session.get,
+                              self.TEST_URL, status_code_retries=retries,
+                              retriable_status_codes=[503, 409])
+
+            self.assertEqual(retries, m.call_count)
+            # The interval maxes out at 60
+            m.assert_called_with(60.0)
+
+        # we count retries so there will be one initial request + 3 retries
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
     def test_uses_tcp_keepalive_by_default(self):
         session = client_session.Session()
         requests_session = session.session
@@ -542,6 +580,118 @@ class SessionTests(utils.TestCase):
                    'value': 'new_name'}])
         self.assertContentTypeIs('application/json-patch+json')
 
+    def test_api_sig_error_message_single(self):
+        title = 'this error is bogus!'
+        detail = 'it is a totally made up error'
+        error_message = {
+            'errors': [
+                {
+                    'request_id': uuid.uuid4().hex,
+                    'code': 'phoney.bologna.error',
+                    'status': 500,
+                    'title': title,
+                    'detail': detail,
+                    'links': [
+                        {
+                            'rel': 'help',
+                            'href': 'https://openstack.org'
+                        }
+                    ]
+                }
+            ]
+        }
+        payload = json.dumps(error_message)
+        self.stub_url('GET', status_code=9000, text=payload,
+                      headers={'Content-Type': 'application/json'})
+        session = client_session.Session()
+
+        # The exception should contain the information from the error response
+        msg = '{} (HTTP 9000)'.format(title)
+        try:
+            session.get(self.TEST_URL)
+        except exceptions.HttpError as ex:
+            self.assertEqual(ex.message, msg)
+            self.assertEqual(ex.details, detail)
+
+    def test_api_sig_error_message_multiple(self):
+        title = 'this error is the first error!'
+        detail = 'it is a totally made up error'
+        error_message = {
+            'errors': [
+                {
+                    'request_id': uuid.uuid4().hex,
+                    'code': 'phoney.bologna.error',
+                    'status': 500,
+                    'title': title,
+                    'detail': detail,
+                    'links': [
+                        {
+                            'rel': 'help',
+                            'href': 'https://openstack.org'
+                        }
+                    ]
+                },
+                {
+                    'request_id': uuid.uuid4().hex,
+                    'code': 'phoney.bologna.error',
+                    'status': 500,
+                    'title': 'some other error',
+                    'detail': detail,
+                    'links': [
+                        {
+                            'rel': 'help',
+                            'href': 'https://openstack.org'
+                        }
+                    ]
+                }
+            ]
+        }
+        payload = json.dumps(error_message)
+        self.stub_url('GET', status_code=9000, text=payload,
+                      headers={'Content-Type': 'application/json'})
+        session = client_session.Session()
+
+        # The exception should contain the information from the error response
+        msg = ('Multiple error responses, showing first only: {} (HTTP 9000)'
+               .format(title))
+        try:
+            session.get(self.TEST_URL)
+        except exceptions.HttpError as ex:
+            self.assertEqual(ex.message, msg)
+            self.assertEqual(ex.details, detail)
+
+    def test_api_sig_error_message_empty(self):
+        error_message = {
+            'errors': [
+            ]
+        }
+        payload = json.dumps(error_message)
+        self.stub_url('GET', status_code=9000, text=payload,
+                      headers={'Content-Type': 'application/json'})
+        session = client_session.Session()
+
+        # The exception should contain the information from the error response
+        msg = 'HTTP Error (HTTP 9000)'
+
+        try:
+            session.get(self.TEST_URL)
+        except exceptions.HttpError as ex:
+            self.assertEqual(ex.message, msg)
+            self.assertIsNone(ex.details)
+
+    def test_error_message_unknown_schema(self):
+        error_message = 'Uh oh, things went bad!'
+        payload = json.dumps(error_message)
+        self.stub_url('GET', status_code=9000, text=payload,
+                      headers={'Content-Type': 'application/json'})
+        session = client_session.Session()
+
+        msg = 'Unrecognized schema in response body. (HTTP 9000)'
+        try:
+            session.get(self.TEST_URL)
+        except exceptions.HttpError as ex:
+            self.assertEqual(ex.message, msg)
+
 
 class RedirectTests(utils.TestCase):
 
@@ -1146,19 +1296,25 @@ class AdapterTest(utils.TestCase):
         self.assertRequestHeaderEqual('User-Agent', self.USER_AGENT)
 
     def test_setting_global_id_on_request(self):
-        global_id = "req-%s" % uuid.uuid4()
+        global_id_adpt = "req-%s" % uuid.uuid4()
+        global_id_req = "req-%s" % uuid.uuid4()
         response = uuid.uuid4().hex
         self.stub_url('GET', text=response)
-        adpt = adapter.Adapter(client_session.Session(),
-                               auth=CalledAuthPlugin(),
-                               service_type=self.SERVICE_TYPE,
-                               service_name=self.SERVICE_NAME,
-                               interface=self.INTERFACE,
-                               region_name=self.REGION_NAME,
-                               user_agent=self.USER_AGENT,
-                               version=self.VERSION,
-                               allow=self.ALLOW,
-                               global_request_id=global_id)
+
+        def mk_adpt(**kwargs):
+            return adapter.Adapter(client_session.Session(),
+                                   auth=CalledAuthPlugin(),
+                                   service_type=self.SERVICE_TYPE,
+                                   service_name=self.SERVICE_NAME,
+                                   interface=self.INTERFACE,
+                                   region_name=self.REGION_NAME,
+                                   user_agent=self.USER_AGENT,
+                                   version=self.VERSION,
+                                   allow=self.ALLOW,
+                                   **kwargs)
+
+        # No global_request_id
+        adpt = mk_adpt()
         resp = adpt.get('/')
         self.assertEqual(resp.text, response)
 
@@ -1166,7 +1322,21 @@ class AdapterTest(utils.TestCase):
         self.assertEqual(self.ALLOW,
                          adpt.auth.endpoint_arguments['allow'])
         self.assertTrue(adpt.auth.get_token_called)
-        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id)
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', None)
+
+        # global_request_id only on the request
+        adpt.get('/', global_request_id=global_id_req)
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id_req)
+
+        # global_request_id only on the adapter
+        adpt = mk_adpt(global_request_id=global_id_adpt)
+        adpt.get('/')
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id_adpt)
+
+        # global_request_id on the adapter *and* the request (the request takes
+        # precedence)
+        adpt.get('/', global_request_id=global_id_req)
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id_req)
 
     def test_setting_variables_on_get_endpoint(self):
         adpt = self._create_loaded_adapter()
